from __future__ import annotations

from .config import setup
from .convert import DefaultConverter
from .fields import ForeignKey, Related
from .schema import ModelSchema

__all__ = (
    "DefaultConverter",
    "ForeignKey",
    "ModelSchema",
    "Related",
    "setup",
)



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Dict, Optional, Type

from marshmallow import Schema, fields

if TYPE_CHECKING:
    import peewee as pw


class Related(fields.Nested):
    def __init__(
        self,
        nested: Optional[Type[Schema]] = None,
        meta: Optional[Dict[str, Any]] = None,
        **kwargs,
    ):
        self.field = None
        self.meta = meta or {}
        super(Related, self).__init__(nested, **kwargs)  # type: ignore[arg-type]

    def init_model(self, model: pw.Model, name: str):
        from .schema import ModelSchema

        field = model._meta.fields.get(name)  # type: ignore[]
        if not field:
            field = getattr(model, name).field
            self.many = True
            rel_model = field.model
        else:
            rel_model = field.rel_model

        self.field = field
        self.attribute = self.attribute or name
        self.meta["model"] = rel_model
        meta = type("Meta", (), self.meta)
        self.nested = type("Schema", (ModelSchema,), {"Meta": meta})

    def _deserialize(self, value, attr, data, **_):
        if self.field is None:
            raise RuntimeError("Init model first.")

        if not isinstance(value, dict):
            return self.field.rel_field.python_value(value)

        return super(Related, self)._deserialize(value, attr, data)


class ForeignKey(fields.Raw):
    string_keys = False

    def _bind_to_schema(self, field_name, schema):
        self.string_keys = schema.opts.string_keys
        super()._bind_to_schema(field_name, schema)

    def get_value(self, obj: pw.Model, *_, **__) -> Any:
        """Return the value for a given key from an object."""
        value = obj.__data__.get(self.attribute)
        if value is not None and self.string_keys:
            return str(value)
        return value



from __future__ import annotations

from typing import Final

DEFAULTS: Final = {
    "dump_only_pk": True,
    "string_keys": True,
    "id_keys": False,
}


def setup(**options):
    """Setup marshmallow-peewee defaults."""
    DEFAULTS.update(options)



from __future__ import annotations

from collections import OrderedDict
from typing import TYPE_CHECKING, List, Optional, Type, cast, overload

import peewee as pw
from marshmallow import ValidationError, fields
from marshmallow import validate as ma_validate

from .fields import ForeignKey

if TYPE_CHECKING:
    from collections.abc import Callable

    from .schema import SchemaOpts
    from .types import TFieldMappingList


class ModelConverterMeta(type):
    """Metaclass for ModelConverter."""

    def __new__(cls, name, bases, attrs):
        kls = super().__new__(cls, name, bases, attrs)
        kls.TYPE_MAPPING = list(kls.TYPE_MAPPING)  # type: ignore[attr-defined]
        return kls


class DefaultConverter(metaclass=ModelConverterMeta):
    """Convert Peewee model to Marshmallow schema."""

    TYPE_MAPPING: TFieldMappingList = []

    def __init__(self, opts: SchemaOpts):
        self.opts = opts

    def get_fields(self, model: pw.Model) -> OrderedDict:
        result = OrderedDict()
        meta = cast(pw.Metadata, model._meta)  # type: ignore[]
        for field in meta.sorted_fields:
            name = field.name
            if self.opts.id_keys and isinstance(field, pw.ForeignKeyField):
                name = field.column_name

            ma_field = self.convert(field)
            if ma_field:
                result[name] = ma_field

        return result

    @overload
    @classmethod
    def register(cls, field: Type[pw.Field]) -> Callable[[Callable], Callable]:
        ...

    @overload
    @classmethod
    def register(cls, field: Type[pw.Field], ma_field: Type[fields.Field]) -> None:
        ...

    @classmethod
    def register(
        cls,
        field: Type[pw.Field],
        ma_field: Optional[Type[fields.Field]] = None,
    ) -> Callable[[Callable], Callable] | None:
        if ma_field is None:

            def wrapper(fn):
                cls.TYPE_MAPPING.insert(0, (field, fn))
                return fn

            return wrapper

        builder = generate_builder(ma_field)
        cls.register(field)(builder)

        return None

    def convert(self, field: pw.Field) -> fields.Field:
        params = {
            "allow_none": field.null,
            "attribute": field.name,
            "required": not field.null and field.default is None,
            "validate": [convert_value_validate(field.db_value)],
        }

        if field.null:
            params["load_default"] = None

        if field.default is not None:
            params["load_default"] = field.default

        if field.choices is not None:
            choices = []
            labels = []
            for c in field.choices:
                choices.append(c[0])
                labels.append(c[1])
            params["validate"].append(ma_validate.OneOf(choices, labels))

        if field.help_text:
            params["metadata"] = {"description": field.help_text}

        # use first "known" field class from field class mro
        # so that extended field classes get converted correctly
        for cls in field.__class__.__mro__:
            for pw_field, builder in self.TYPE_MAPPING:
                if cls is pw_field:
                    return builder(field, self.opts, **params)

        return DEFAULT_BUILDER(field, self.opts, **params)


def generate_builder(ma_field_cls: Type[fields.Field]) -> Callable:
    """Generate builder function for given marshmallow field."""

    def builder(_: pw.Field, __: SchemaOpts, **params) -> fields.Field:
        return ma_field_cls(**params)

    return builder


DEFAULT_BUILDER = generate_builder(fields.Raw)


DefaultConverter.register(pw.IntegerField, ma_field=fields.Integer)
DefaultConverter.register(pw.DateTimeField, ma_field=fields.DateTime)
DefaultConverter.register(pw.DateField, ma_field=fields.Date)
DefaultConverter.register(pw.TextField, ma_field=fields.String)
DefaultConverter.register(pw.ForeignKeyField, ma_field=ForeignKey)
DefaultConverter.register(pw.DeferredForeignKey, ma_field=ForeignKey)
DefaultConverter.register(pw.FloatField, ma_field=fields.Float)
DefaultConverter.register(pw.DecimalField, ma_field=fields.Decimal)
DefaultConverter.register(pw.TimeField, ma_field=fields.Time)
DefaultConverter.register(pw.BigIntegerField, ma_field=fields.Integer)
DefaultConverter.register(pw.SmallIntegerField, ma_field=fields.Integer)
DefaultConverter.register(pw.DoubleField, ma_field=fields.Float)
DefaultConverter.register(pw.FixedCharField, ma_field=fields.String)
DefaultConverter.register(pw.UUIDField, ma_field=fields.UUID)


@DefaultConverter.register(pw.AutoField)  # type: ignore[]
def convert_autofield(_: pw.Field, opts: SchemaOpts, **params) -> fields.Field:
    ftype = fields.String if opts.string_keys else fields.Integer
    params["required"] = False
    return ftype(dump_only=opts.dump_only_pk, **params)


@DefaultConverter.register(pw.CharField)  # type: ignore[]
def convert_charfield(
    field: pw.CharField,
    _: SchemaOpts,
    *,
    validate: Optional[List] = None,
    **params,
) -> fields.Field:
    if validate is None:
        validate = []

    validate += [ma_validate.Length(max=field.max_length)]
    return fields.String(validate=validate, **params)


@DefaultConverter.register(pw.BooleanField)  # type: ignore[]
def convert_booleanfield(
    _: pw.BooleanField,
    __: SchemaOpts,
    **params,
) -> fields.Field:
    return fields.Boolean(**params)


def convert_value_validate(converter: Callable) -> Callable:
    def validator(value):
        try:
            converter(value)
        except Exception as exc:  # noqa: BLE001
            raise ValidationError(str(exc)) from exc

    return validator


# ruff: noqa: N802, N815, ARG002



from __future__ import annotations

from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Generic,
    Iterable,
    List,
    Literal,
    Mapping,
    Optional,
    Type,
    Union,
    overload,
)

import marshmallow as ma
import peewee as pw
from marshmallow import schema

from .config import DEFAULTS
from .convert import DefaultConverter
from .fields import Related
from .types import TVModel


class SchemaOpts(ma.SchemaOpts, Generic[TVModel]):
    model: Optional[Type[TVModel]]
    dump_only_pk: bool
    string_keys: bool
    id_keys: bool
    model_converter: Type[DefaultConverter]

    def __init__(self, meta, **kwargs):
        super(SchemaOpts, self).__init__(meta, **kwargs)
        self.model = getattr(meta, "model", None)
        self.dump_only_pk = getattr(meta, "dump_only_pk", DEFAULTS["dump_only_pk"])
        self.string_keys = getattr(meta, "string_keys", DEFAULTS["string_keys"])
        self.id_keys = getattr(meta, "id_keys", DEFAULTS["id_keys"])

        if self.model and not issubclass(self.model, pw.Model):
            raise ValueError("`model` must be a subclass of peewee.Model")

        self.model_converter = getattr(meta, "model_converter", DefaultConverter)


INHERITANCE_OPTIONS = (
    "model",
    "model_converter",
    "dump_only_pk",
    "string_keys",
    # Basic options
    "datetimeformat",
    "dateformat",
    "timeformat",
    "unknown",
)


class SchemaMeta(schema.SchemaMeta):
    def __new__(cls, name, bases, attrs):
        """Support inheritance for model and model_converter Meta options."""
        if "Meta" in attrs and bases:
            meta = attrs["Meta"]
            base_meta = getattr(bases[0], "Meta", None)
            for option in INHERITANCE_OPTIONS:
                if hasattr(meta, option) or not hasattr(base_meta, option):
                    continue
                setattr(meta, option, getattr(base_meta, option))

        return super(SchemaMeta, cls).__new__(cls, name, bases, attrs)

    @classmethod
    def get_declared_fields(cls, klass, cls_fields, inherited_fields, dict_cls):
        opts: SchemaOpts = klass.opts
        base_fields = super(SchemaMeta, cls).get_declared_fields(
            klass, cls_fields, inherited_fields, dict_cls
        )
        declared_fields = dict_cls()
        model = getattr(opts, "model", None)
        if model is not None:
            for name, field in base_fields.items():
                if isinstance(field, Related) and field.nested is None:
                    field.init_model(model, name)

            converter = opts.model_converter(opts=opts)
            declared_fields.update(converter.get_fields(model))
        declared_fields.update(base_fields)
        return declared_fields


class ModelSchema(ma.Schema, Generic[TVModel], metaclass=SchemaMeta):
    OPTIONS_CLASS = SchemaOpts

    opts: SchemaOpts[TVModel]

    def __init__(self, instance: Optional[TVModel] = None, **kwargs):
        self.instance = instance
        super(ModelSchema, self).__init__(**kwargs)

    @overload  # type: ignore[override]
    def load(self, data, *, many: Optional[Literal[False]] = None, **kwargs) -> TVModel:
        ...

    @overload
    def load(
        self, data, *, many: Optional[Literal[True]] = None, **kwargs
    ) -> List[TVModel]:
        ...

    def load(
        self,
        data: Union[Mapping[str, Any], Iterable[Mapping[str, Any]]],
        *,
        instance: Optional[TVModel] = None,
        **kwargs,
    ):
        self.instance = instance or self.instance
        return super().load(data, **kwargs)

    @ma.post_load
    def make_instance(self, data: Dict[str, Any], **params) -> Union[Dict, TVModel]:
        """Build object from data."""
        if not self.opts.model:
            return data

        if self.instance is None:
            return self.opts.model(**data)

        for key, value in data.items():
            setattr(self.instance, key, value)

        return self.instance

    if TYPE_CHECKING:

        @overload  # type: ignore[override]
        def dump(self, obj) -> Dict[str, Any]:
            ...

        @overload
        def dump(self, obj, *, many: Literal[False]) -> Dict[str, Any]:
            ...

        @overload
        def dump(self, obj, *, many: Literal[True]) -> List[Dict[str, Any]]:
            ...

        def dump(
            self, obj: Union[TVModel, Iterable[TVModel]], *, many: Optional[bool] = None
        ) -> Union[Dict[str, Any], List[Dict[str, Any]]]:
            ...



from __future__ import annotations

from typing import Callable, List, Tuple, Type, TypeVar

import marshmallow as ma
import peewee as pw

TVModel = TypeVar("TVModel", bound=pw.Model)
TFieldMappingList = List[Tuple[Type[pw.Field], Callable[..., ma.fields.Field]]]
